/*
 * Jalview - A Sequence Alignment Editor and Viewer
 * Copyright (C) 2005 AM Waterhouse, J Procter, G Barton, M Clamp, S Searle
 *
 * This program is free software; you can redistribute it and/or
 * modify it under the terms of the GNU General Public License
 * as published by the Free Software Foundation; either version 2
 * of the License, or (at your option) any later version.
 *
 * This program is distributed in the hope that it will be useful,
 * but WITHOUT ANY WARRANTY; without even the implied warranty of
 * MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
 * GNU General Public License for more details.
 *
 * You should have received a copy of the GNU General Public License
 * along with this program; if not, write to the Free Software
 * Foundation, Inc., 51 Franklin Street, Fifth Floor, Boston, MA  02110-1301, USA
 */

package jalview.appletgui;

import java.awt.*;

import jalview.analysis.*;
import jalview.datamodel.*;

public class SeqCanvas
    extends Panel
{
  FeatureRenderer fr;
  SequenceRenderer sr;
  Image img;
  Graphics gg;
  int imgWidth;
  int imgHeight;

  AlignViewport av;

  boolean showScores = false;
  boolean displaySearch = false;
  int[] searchResults = null;

  int chunkHeight;
  int chunkWidth;

  boolean fastPaint = false;

  public SeqCanvas(AlignViewport av)
  {
    this.av = av;
    fr = new FeatureRenderer(av);
    sr = new SequenceRenderer(av);
    PaintRefresher.Register(this, av.alignment);

  }

  void drawNorthScale(Graphics g, int startx, int endx, int ypos)
  {
    int scalestartx = startx - startx % 10 + 10;

    g.setColor(Color.black);

    // NORTH SCALE
    for (int i = scalestartx; i < endx; i += 10)
    {
      String string = String.valueOf(i);
      g.drawString(string, (i - startx - 1) * av.charWidth,
                   ypos - av.charHeight / 2);

      g.drawLine( (i - startx - 1) * av.charWidth + av.charWidth / 2,
                 ypos + 2 - av.charHeight / 2,
                 (i - startx - 1) * av.charWidth + av.charWidth / 2, ypos - 2);

    }
  }

  void drawWestScale(Graphics g, int startx, int endx, int ypos)
  {
    FontMetrics fm = getFontMetrics(av.getFont());
    ypos += av.charHeight;
    // EAST SCALE
    for (int i = 0; i < av.alignment.getHeight(); i++)
    {
      SequenceI seq = av.alignment.getSequenceAt(i);
      int index = startx;
      int value = -1;
      while (index < endx)
      {
        if (jalview.util.Comparison.isGap(seq.getCharAt(index)))
        {
          index++;
          continue;
        }

        value = av.alignment.getSequenceAt(i).findPosition(index);
        break;
      }
      if (value != -1)
      {
        int x = LABEL_WEST - fm.stringWidth(value + "");
        g.drawString(value + "", x,
                     ypos + i * av.charHeight - av.charHeight / 5);
      }
    }
  }

  void drawEastScale(Graphics g, int startx, int endx, int ypos)
  {
    ypos += av.charHeight;
    // EAST SCALE
    for (int i = 0; i < av.alignment.getHeight(); i++)
    {
      SequenceI seq = av.alignment.getSequenceAt(i);
      int index = endx;
      int value = -1;
      while (index > startx)
      {
        if (jalview.util.Comparison.isGap(seq.getCharAt(index)))
        {
          index--;
          continue;
        }

        value = av.alignment.getSequenceAt(i).findPosition(index);
        break;
      }
      if (value != -1)
      {
        g.drawString(value + "", 0,
                     ypos + i * av.charHeight - av.charHeight / 5);
      }
    }

  }

  public void fastPaint(int horizontal, int vertical)
  {
    if (horizontal == 0 && vertical == 0 || gg == null)
    {
      return;
    }

    gg.copyArea(0, 0, imgWidth, imgHeight, -horizontal * av.charWidth,
                -vertical * av.charHeight);

    int sr = av.startRes, er = av.endRes, ss = av.startSeq, es = av.endSeq,
        transX = 0, transY = 0;
    if (horizontal > 0) // scrollbar pulled right, image to the left
    {
      transX = (er - sr - horizontal) * av.charWidth;
      sr = er - horizontal;
    }
    else if (horizontal < 0)
    {
      er = sr - horizontal;
    }

    else if (vertical > 0) // scroll down
    {
      ss = es - vertical;
      if (ss < av.startSeq) // ie scrolling too fast, more than a page at a time
      {
        ss = av.startSeq;
      }
      else
      {
        transY = imgHeight - vertical * av.charHeight;
      }
    }
    else if (vertical < 0)
    {
      es = ss - vertical;
      if (es > av.endSeq)
      {
        es = av.endSeq;
      }
    }

    gg.translate(transX, transY);

    gg.setColor(Color.white);
    gg.fillRect(0, 0, (er - sr + 1) * av.charWidth, (es - ss) * av.charHeight);
    drawPanel(gg, sr, er, ss, es, sr, ss, 0);
    gg.translate( -transX, -transY);

    fastPaint = true;
    repaint();

  }

  /**
   * Definitions of startx and endx (hopefully):
   * SMJS This is what I'm working towards!
   *   startx is the first residue (starting at 0) to display.
   *   endx   is the last residue to display (starting at 0).
   *   starty is the first sequence to display (starting at 0).
   *   endy   is the last sequence to display (starting at 0).
   * NOTE 1: The av limits are set in setFont in this class and
   * in the adjustment listener in SeqPanel when the scrollbars move.
   */
  public void update(Graphics g)
  {
    paint(g);
  }

  public void paint(Graphics g)
  {
    if (fastPaint)
    {
      g.drawImage(img, 0, 0, this);
      fastPaint = false;
      return;
    }

    // this draws the whole of the alignment
    imgWidth = this.getSize().width;
    imgHeight = this.getSize().height;

    imgWidth -= imgWidth % av.charWidth;
    imgHeight -= imgHeight % av.charHeight;

    if (imgWidth < 1 || imgHeight < 1)
    {
      return;
    }

    if (img == null || imgWidth != img.getWidth(this) ||
        imgHeight != img.getHeight(this))
    {
      img = createImage(imgWidth, imgHeight);
      gg = img.getGraphics();
      gg.setFont(av.getFont());
    }

    gg.setColor(Color.white);
    gg.fillRect(0, 0, imgWidth, imgHeight);

    chunkWidth = getWrappedCanvasWidth(getSize().width);
    chunkHeight = (av.getAlignment().getHeight() + 2) * av.charHeight;

    av.setChunkHeight(chunkHeight);
    av.setChunkWidth(chunkWidth);

    if (av.getWrapAlignment())
    {
      drawWrappedPanel(gg, getSize().width, getSize().height, av.startRes);
    }
    else
    {
      drawPanel(gg, av.startRes, av.endRes, av.startSeq, av.endSeq, av.startRes,
                av.startSeq, 0);
    }

    g.drawImage(img, 0, 0, this);

  }

  int LABEL_WEST, LABEL_EAST;
  public int getWrappedCanvasWidth(int cwidth)
  {
    FontMetrics fm = getFontMetrics(av.getFont());

    LABEL_EAST = 0;
    LABEL_WEST = 0;

    if (av.scaleRightWrapped)
    {
      LABEL_EAST = fm.stringWidth(av.alignment.getWidth() + "000");
    }

    if (av.scaleLeftWrapped)
    {
      LABEL_WEST = fm.stringWidth(av.alignment.getWidth() + "");
    }

    return (cwidth - LABEL_EAST - LABEL_WEST) / av.charWidth;
  }

  public void drawWrappedPanel(Graphics g, int canvasWidth, int canvasHeight,
                               int startRes)
  {
    AlignmentI al = av.getAlignment();

    FontMetrics fm = getFontMetrics(av.getFont());

    int LABEL_EAST = 0;
    if (av.scaleRightWrapped)
    {
      LABEL_EAST = fm.stringWidth(al.getWidth() + "000");
    }
    int LABEL_WEST = 0;
    if (av.scaleLeftWrapped)
    {
      LABEL_WEST = fm.stringWidth(al.getWidth() + "0");
    }

    int cWidth = (canvasWidth - LABEL_EAST - LABEL_WEST) / av.charWidth;
    int cHeight = (av.getAlignment().getHeight() + 2) * av.charHeight;

    av.endRes = av.startRes + cWidth;

    int endx = startRes + cWidth - 1;
    int ypos = 2 * av.charHeight;

    while (ypos <= canvasHeight && startRes < av.alignment.getWidth())
    {
      g.setColor(Color.black);

      if (av.scaleLeftWrapped)
      {
        drawWestScale(g, startRes, endx, ypos);
      }

      if (av.scaleRightWrapped)
      {
        g.translate(canvasWidth - LABEL_EAST + av.charWidth, 0);
        drawEastScale(g, startRes, endx, ypos);
        g.translate( - (canvasWidth - LABEL_EAST + av.charWidth), 0);
      }

      g.translate(LABEL_WEST, 0);
      if (av.scaleAboveWrapped)
      {
        drawNorthScale(g, startRes, endx, ypos);
      }

      // When printing we have an extra clipped region,
      // the Printable page which we need to account for here
      Shape clip = g.getClip();
      if (clip == null)
      {
        g.setClip(0, 0, cWidth * av.charWidth, canvasHeight);
      }
      else
      {
        g.setClip(0,
                  (int) clip.getBounds().y,
                  cWidth * av.charWidth,
                  (int) clip.getBounds().height
            );
      }

      drawPanel(g, startRes, endx, 0, al.getHeight(), startRes, 0, ypos);
      g.setClip(clip);
      g.translate( -LABEL_WEST, 0);

      ypos += cHeight;
      startRes += cWidth;
      endx = startRes + cWidth - 1;

      if (endx > al.getWidth())
      {
        endx = al.getWidth();
      }
    }

  }

  synchronized public void drawPanel(Graphics g, int x1, int x2, int y1, int y2,
                                     int startx, int starty, int offset)
  {

    g.setFont(av.getFont());
    sr.renderGaps(av.renderGaps);

    SequenceI nextSeq;

    /// First draw the sequences
    /////////////////////////////
    for (int i = y1; i < y2; i++)
    {
      nextSeq = av.alignment.getSequenceAt(i);

      sr.drawSequence(g, nextSeq, av.alignment.findAllGroups(nextSeq), x1, x2,
                      (x1 - startx) * av.charWidth,
                      offset +
                      (i-starty)*av.charHeight,
                      av.charWidth, av.charHeight);

      if (av.showSequenceFeatures)
      {
        fr.drawSequence(g, nextSeq, av.alignment.findAllGroups(nextSeq), x1, x2,
                        (x1 - startx) * av.charWidth,
                        offset +
                        (i-starty)*av.charHeight,
                        av.charWidth, av.charHeight);
      }
    }
    //
    /////////////////////////////////////

    // Now outline any areas if necessary
    /////////////////////////////////////
    SequenceGroup group = av.getSelectionGroup();
    java.util.Vector groups = av.alignment.getGroups();

    int sx = -1, sy = -1, ex = -1;
    int groupIndex = -1;
    if (group == null && groups.size() > 0)
    {
      group = (SequenceGroup) groups.elementAt(0);
      groupIndex = 0;
    }

    if (group != null)
    {
      do
      {
        int oldY = -1;
        int i = 0;
        boolean inGroup = false;
        int top = -1, bottom = -1;
        for (i = y1; i < y2; i++)
        {
          sx = (group.getStartRes() - startx) * av.charWidth;
          sy = offset + (i-starty)*av.charHeight;
          ex = (group.getEndRes() + 1 - group.getStartRes()) * av.charWidth - 1;

          if (sx < getSize().width
              && ex > 0
              && group.sequences.contains(av.alignment.getSequenceAt(i)))
          {
            if (bottom == -1 && (i == av.alignment.getHeight() - 1 ||
                                 !group.sequences.contains(av.alignment.
                getSequenceAt(i + 1))))
            {
              bottom = sy + av.charHeight;
            }

            if (!inGroup)
            {
              if (top == -1 && i == 0 ||
                  !group.sequences.contains(av.alignment.getSequenceAt(i - 1)))
              {
                top = sy;
              }

              oldY = sy;
              inGroup = true;
              if (group == av.getSelectionGroup())
              {
                g.setColor(new Color(255, 0, 0));
              }
              else
              {
                g.setColor(group.getOutlineColour());
              }
            }
          }
          else
          {
            if (inGroup)
            {
              g.drawLine(sx, oldY, sx, sy);
              g.drawLine(sx + ex, oldY, sx + ex, sy);

              if (top != -1)
              {
                g.drawLine(sx, top, sx + ex, top);
                top = -1;
              }
              if (bottom != -1)
              {
                g.drawLine(sx, bottom, sx + ex, bottom);
                bottom = -1;
              }

              inGroup = false;
            }
          }
        }

        if (inGroup)
        {

          if (top != -1)
          {
            g.drawLine(sx, top, sx + ex, top);
            top = -1;
          }
          if (bottom != -1)
          {
            g.drawLine(sx, bottom - 1, sx + ex, bottom - 1);
            bottom = -1;

          }
          sy = offset + (i-starty)*av.charHeight;
          g.drawLine(sx, oldY, sx, sy);
          g.drawLine(sx + ex, oldY, sx + ex, sy);
          inGroup = false;
        }
        groupIndex++;
        if (groupIndex >= groups.size())
        {
          break;
        }

        group = (SequenceGroup) groups.elementAt(groupIndex);

      }
      while (groupIndex < groups.size());
    }

    /// Highlight search Results once all sequences have been drawn
    //////////////////////////////////////////////////////////
    if (displaySearch)
    {
      for (int r = 0; r < searchResults.length; r += 3)
      {
        int searchSeq = searchResults[r];

        if (searchSeq >= y1 && searchSeq < y2)
        {
          SequenceI seq = av.getAlignment().getSequenceAt(searchSeq);

          int searchStart = seq.findIndex(searchResults[r + 1]) - 1;
          int searchEnd = seq.findIndex(searchResults[r + 2]) - 1;

          SequenceRenderer ssr = (SequenceRenderer) sr;
          if (searchStart < x1)
          {
            searchStart = x1;
          }
          if (searchEnd > x2)
          {
            searchEnd = x2;
          }

          ssr.drawHighlightedText(seq,
                                  searchStart,
                                  searchEnd,
                                  (searchStart - startx) * av.charWidth,
                                  offset +
                                  (searchSeq-starty)*av.charHeight,
                                  av.charWidth,
                                  av.charHeight);
        }
      }
    }

  }

  public void highlightSearchResults(int[] results)
  {
    // results are in the order sequence, startRes, endRes
    if (results == null)
    {
      displaySearch = false;
    }
    else
    {
      displaySearch = true;
    }

    searchResults = results;

    repaint();
  }

}
